#include "openssl/bio.h"
#include "openssl/ssl.h"
#include "openssl/err.h"

#define CERT_PATH "packz-certificate.pem"
#define PKEY_PATH "packz-private-key.pem"

#define CHK_NULL(x) if ((x)==NULL) exit (1)
#define CHK_SSL(err) if ((err)==-1) { ERR_print_errors_fp(stderr); exit(2); }

/*remember to clean the shit*/

BIO* ssl_server_wait(){
  SSL_CTX *ctx;
  SSL *ssl;
  BIO *bio, *abio, *out;//, *sbio;
  fprintf(stderr, "Attempting to create SSL context... \r");
  ctx = SSL_CTX_new(SSLv23_server_method());
  if(ctx == NULL){
    ERR_print_errors_fp(stderr);
    printf("Failed. Aborting.\n");
    return 0;
  }

  fprintf(stderr, "Loading certificates...\r");
  //SSL_CTX_set_default_passwd_cb(ctx, callback);
  if(!SSL_CTX_use_certificate_file(ctx, CERT_PATH, SSL_FILETYPE_PEM))
  {
    ERR_print_errors_fp(stdout);
    SSL_CTX_free(ctx);
    return 0;
  }
  if(!SSL_CTX_use_PrivateKey_file(ctx, PKEY_PATH, SSL_FILETYPE_PEM))
  {
    ERR_print_errors_fp(stdout);
    SSL_CTX_free(ctx);
    return 0;
  }

  fprintf(stderr, "Attempting to create BIO object... \r");
  bio = BIO_new_ssl(ctx, 0);
  if(bio == NULL){
    printf("Failed. Aborting.\n");
    ERR_print_errors_fp(stdout);
    SSL_CTX_free(ctx);
    return 0;
  }

  fprintf(stderr, "Attempting to set up BIO for SSL...\r");
  BIO_get_ssl(bio, &ssl);
  SSL_set_mode(ssl, SSL_MODE_AUTO_RETRY);

  abio = BIO_new_accept("1137");
  BIO_set_accept_bios(abio, bio);

  fprintf(stderr, "Waiting for incoming connection...\r");

  if(BIO_do_accept(abio) <= 0){
    ERR_print_errors_fp(stdout);
    SSL_CTX_free(ctx);
    BIO_free_all(bio);
    BIO_free_all(abio);
    return NULL;
  }

  if(BIO_do_accept(abio) <= 0){
    ERR_print_errors_fp(stdout);
    SSL_CTX_free(ctx);
    BIO_free_all(bio);
    BIO_free_all(abio);
    return NULL;
  }

  out = BIO_pop(abio);

  if(BIO_do_handshake(out) <= 0){
    printf("Handshake failed.\n");
    ERR_print_errors_fp(stdout);
    SSL_CTX_free(ctx);
    BIO_free_all(bio);
    BIO_free_all(abio);
    return NULL;
  }

	return out;
}

BIO* ssl_client_connect(char* hostport){
  BIO * bio;
  SSL * ssl;
  SSL_CTX * ctx;
  char*    str;
  X509*    server_cert;
  /* Set up the SSL context */
  ctx = SSL_CTX_new(SSLv23_client_method());
  if(!ctx){
    fprintf(stderr, "* error creating ctx\n");
    return NULL;
  }

  /* Load the trust store */

  if(!SSL_CTX_load_verify_locations(ctx, CERT_PATH, NULL)){
    fprintf(stderr, "Error loading trust store\n");
    ERR_print_errors_fp(stderr);
    SSL_CTX_free(ctx);
    return NULL;
  }

  /* Setup the connection */
  bio = BIO_new_ssl_connect(ctx);

  /* Set the SSL_MODE_AUTO_RETRY flag */
  BIO_get_ssl(bio, & ssl);
  SSL_set_mode(ssl, SSL_MODE_AUTO_RETRY);

  /* Create and setup the connection */
  BIO_set_conn_hostname(bio, hostport);

  if(BIO_do_connect(bio) <= 0){
    fprintf(stderr, "Error attempting to connect\n");
    ERR_print_errors_fp(stderr);
    BIO_free_all(bio);
    SSL_CTX_free(ctx);
    return NULL;
  }

  /* Check the certificate */
  if( SSL_get_verify_result(ssl) != X509_V_OK ){
    fprintf(stderr,
        "Certificate verification error: %ld\n", SSL_get_verify_result(ssl));
    ERR_print_errors_fp(stderr);
    BIO_free_all(bio);
    SSL_CTX_free(ctx);
    return NULL;
  }

  fprintf (stderr, "* SSL connection using %s\n", SSL_get_cipher (ssl));
  server_cert = SSL_get_peer_certificate (ssl);
  CHK_NULL(server_cert);
  fprintf (stderr, "* Server certificate:\n");

  str = X509_NAME_oneline (X509_get_subject_name (server_cert),0,0);
  CHK_NULL(str);
  fprintf (stderr, "* subject: %s\n", str);
	return bio;
}

int io_loop(BIO* out){
	int fd;
	fd = BIO_get_fd(out, 0);
	if( fd < 0 ){
		fprintf(stderr, "* error on BIO_get_fd\n");
		return 1;
	}

	fd_set rfd, wfd;
	char msg[1024];
	char outmsg[1024];
	msg[0] = '\0';
	int nbytes;
	while(1){
		FD_ZERO(&rfd);
		FD_ZERO(&wfd);
		if(*msg)/*check socket only on available msg*/
			FD_SET(fd,&wfd);
		FD_SET(0, &rfd);
		FD_SET(fd, &rfd);
		if( select(FD_SETSIZE, &rfd, &wfd, NULL, NULL) < 0 )
			perror("select");

		if( FD_ISSET(0, &rfd) ){
			fgets(msg, 1024, stdin);
		}
    /*
     * select() is not directly reliable about socket 
     * available for reading. With SSL_read() you have
     * to use SSL_pending().
     */
		if( FD_ISSET(fd, &rfd)){// && BIO_pending(out)){
			fprintf(stderr, ">> ");
			nbytes = BIO_read(out, outmsg, 1024);

			if(nbytes < 0)
				ERR_print_errors_fp(stderr);
      else if(nbytes == 0)
        return 1;

			outmsg[nbytes] = '\0';
			fprintf(stderr, "[%d] %s\n", nbytes, outmsg);
		}
		if( FD_ISSET(fd, &wfd) && *msg ){
      fprintf(stderr, "<< ");
      nbytes = BIO_write(out, msg, strlen(msg));
      if( nbytes <= 0 )
				ERR_print_errors_fp(stderr);
      fprintf(stderr, " [%d]\n", nbytes);
			msg[0] = '\0';
		}
	}
  return 0;
}
