/*
 *  http://h71000.www7.hp.com/doc/83final/BA554_90007/ch04s03.html
 *  http://www.rtfm.com/openssl-examples/
 *  openssl genrsa -out privkey.pem 2048
 *  openssl req -new -x509 -key privkey.pem -out cacert.pem -days 1095
 */
#include<stdio.h>
#include<unistd.h>
#include<string.h>
#include<sys/stat.h>
#include<libgen.h>

#include<openssl/bio.h>
#include<openssl/ssl.h>
#include<openssl/err.h>

#include<libtext.h>
#include<openssl-routine.h>
#include<transfer-routine.h>

#define TRSIZE 1024

#define USAGE \
      "usage:" \
      "\tsslpeer share filename\n" \
      "\tsslpeer pull hostname filename\n\n"

static int usage(int narg, char* argv[], char** fname){
  if( narg < 3 ){
    fprintf(stderr, USAGE);
    return 1;
  }

  char* command = argv[1];
  if( strcmp("share", command) == 0 ){
    *fname = argv[2];
  }else
    return 1;

  return 0;
}

int main(int argc, char* argv[]){
  BIO *out;
  char* path,* filename;
  int fdfile;
  int nread, nwrite, ntotalwrite, ntotalsend, run = 1;
  char filenamesize;
  struct stat filestat;
  struct progress* p;

  SSL_library_init();
  SSL_load_error_strings();
  ERR_load_BIO_strings();
  ERR_load_SSL_strings();
  OpenSSL_add_all_algorithms();

  if( usage(argc, argv, &path) != 0 )
    return 1;

  fdfile = get_fd_in(path);
  if( fdfile < 0){
    perror("* error");
    return 1;
  }
  /*reads size*/
  if( fstat(fdfile, &filestat) < 0){
    perror("fstat");
    return 1;
  }

  out = ssl_server_wait();
  if( !out )
    return 1;

  fprintf(stderr, "* connection successfull \n");

  filename = basename(path);
  filenamesize = (char)strlen(filename);
  if( BIO_write(out, (void*)&filenamesize, 1) < 0 )
    ERR_print_errors_fp(stderr);

  if(BIO_puts(out, filename) < 0)
    ERR_print_errors_fp(stderr);

  if(BIO_write(out, &filestat.st_size, 4) < 0)
    ERR_print_errors_fp(stderr);

  p = progress_start(truncate_text_too_long(filename, 40), filestat.st_size);
  char* buf[TRSIZE];
  while(run){
    nread = read(fdfile, buf, TRSIZE);
    if( nread < 0 ){
      perror("* error reading input file");
      run = 0;
      break;
    }else if( nread == 0)
      break;

    ntotalwrite = 0;
    while( nread > ntotalwrite){
      nwrite = BIO_write(out, buf + ntotalwrite, nread - ntotalwrite);
      if(nwrite <= 0){
        perror("* error sending");
        run = 0;
        break;
      }
      ntotalwrite += nwrite;
    }
    ntotalsend += ntotalwrite;
    progress_display(p, ntotalsend);
  }

  //io_loop(out);

  BIO_flush(out);

  return 0;
}

